/*
 * Copyright (c) 2009 Cyrille Berger <cberger@cberger.net>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either 
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library.  If not, see <http://www.gnu.org/licenses/>. */

#include "ShivaRenderWindow.h"

#include <QDockWidget>
#include <QThread>

#include <klocale.h>

#include <OpenShiva/Kernel.h>
#include <QtShiva/KernelParametersWidget.h>
#include <QtGTL/QImageImage.h>
#include <ImageViewer.h>
#include <kdebug.h>
#include <widgets/TransformWidget.h>

class ShivaRenderThread : public QThread
{
  public:
    ShivaRenderThread(const GTLCore::RegionI& region, OpenShiva::Kernel* _kernel, const std::list<const GTLCore::AbstractImage*>& inputImages, const GTLCore::Transform& transform) : m_region(region), m_kernel(_kernel), m_inputImages(inputImages), m_transform(transform)
    {
    }
    void run()
    {
      m_kernel->compile();
      if(m_kernel->isCompiled())
      {
        QtGTL::QImageImage image( QImage( m_region.columns(), m_region.rows(), QImage::Format_ARGB32) );
        m_kernel->evaluatePixels(m_region, m_inputImages, &image, 0, m_transform);
        m_result = image.image();
      }
    }
    QImage result()
    {
      return m_result;
    }
  private:
    QImage m_result;
    GTLCore::RegionI m_region;
    OpenShiva::Kernel* m_kernel;
    std::list<const GTLCore::AbstractImage*> m_inputImages;
    GTLCore::Transform m_transform;
};

ShivaRenderWindow::ShivaRenderWindow() : m_mutex(QMutex::Recursive), m_kernel(0), m_imageViewer(new ImageViewer), m_thread(0), m_restart(false)
{
  setCentralWidget(m_imageViewer);
  
  // Transform docker
  QDockWidget* transformDockerDockWidget = new QDockWidget(i18n("Transform"), this);
  transformDockerDockWidget->setObjectName("ParametersDocker");
  m_transformWidget = new TransformWidget(this);
  connect(m_transformWidget, SIGNAL(transformChanged()), SLOT(render()));
  transformDockerDockWidget->setWidget(m_transformWidget);
  addDockWidget( Qt::RightDockWidgetArea, transformDockerDockWidget );
  
  // Parameters docker
  QDockWidget* parametersDockerDockWidget = new QDockWidget(i18n("Parameters"), this);
  parametersDockerDockWidget->setObjectName("ParametersDocker");
  m_parametersWidget = new QtShiva::KernelParametersWidget( parametersDockerDockWidget );
  connect(m_parametersWidget, SIGNAL(configurationChanged()), SLOT(render()));
  parametersDockerDockWidget->setWidget( m_parametersWidget );
  
  addDockWidget( Qt::RightDockWidgetArea, parametersDockerDockWidget );
}

ShivaRenderWindow::~ShivaRenderWindow()
{
  delete m_kernel;
}

void ShivaRenderWindow::setRegion(const GTLCore::RegionI& region)
{
  m_region = region;
}

void ShivaRenderWindow::render()
{
  QMutexLocker l(&m_mutex);
  if(m_thread)
  {
    m_restart = true;
  } else {
    startRendering();
  }
}

void ShivaRenderWindow::startRendering()
{
  QMutexLocker l(&m_mutex);
  m_restart = false;
  Q_ASSERT(m_thread == 0);
  m_thread = new ShivaRenderThread(m_region, m_kernel, inputImages(), m_transformWidget->transform());
  connect(m_thread, SIGNAL(finished()), SLOT(renderFinished()));
  m_thread->start();
}

void ShivaRenderWindow::renderFinished()
{
  QMutexLocker l(&m_mutex);
  m_imageViewer->setImage(m_thread->result());
  delete m_thread;
  m_thread = 0;
  if(m_restart)
  {
    startRendering();
  }
}

void ShivaRenderWindow::setKernel(OpenShiva::Kernel* kernel)
{
  {
    QMutexLocker l(&m_mutex);
    
    if(m_thread)
    {
      m_thread->wait();
    }
  }
  
  delete m_kernel;
  m_kernel = kernel;
  m_parametersWidget->setKernel( m_kernel );
  render();
}

#include "ShivaRenderWindow.moc"
