/*
 * Copyright (c) 2008 Cyrille Berger <cberger@cberger.net>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either 
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library.  If not, see <http://www.gnu.org/licenses/>. */

#include "gtldesigner.h"

#include "QtGTL/QImageImage.h"

#include <QDockWidget>
#include <QListView>
#include <QDropEvent>
#include <QPainter>
#include <QPrinter>

#include <kconfigdialog.h>
#include <kstatusbar.h>

#include <kaction.h>
#include <kactioncollection.h>
#include <kstandardaction.h>

#include <KLocale>
#include <KMessageBox>

#include <KFileDialog>
#include <ktexteditor/document.h>
#include <ktexteditor/view.h>
#include <ktexteditor/editor.h>
#include <ktexteditor/editorchooser.h>
#include <ktexteditor/configpage.h>
#include <ktexteditor/cursor.h>

#include <GTLCore/PixelDescription.h>
#include <GTLCore/Region.h>
#include <GTLCore/Value.h>

#include <GTLImageIO/ImageDCRegistry.h>
#include <GTLImageIO/ImageDC.h>

#include <OpenShiva/Kernel.h>
#include <OpenCTL/Module.h>
#include <OpenCTL/Program.h>

#include "ImageViewer.h"
#include "ErrorMessagesModel.h"
#include <GTLCore/CompilationMessages.h>
#include "RenderWindow/ShivaRenderWindow.h"

GTLDesigner::GTLDesigner()
    : KXmlGuiWindow(),
      m_view(0),
      m_shivaRenderWindow(0)
{
    // accept dnd
    setAcceptDrops(true);
    KTextEditor::EditorChooser::editor()->readConfig();
    KTextEditor::Editor *editor = KTextEditor::EditorChooser::editor();
    m_document = editor->createDocument(this);
    connect( m_document, SIGNAL(documentNameChanged( KTextEditor::Document *) ), SLOT(updateCaption()) );
    connect( m_document, SIGNAL(textChanged( KTextEditor::Document *) ), SLOT(updateCaption()) );
    
    m_view = qobject_cast<KTextEditor::View*>(m_document->createView(this));
    
    QMetaObject::invokeMethod(m_view, "toggleLineNumbersOn", Qt::DirectConnection );

    // tell the KXmlGuiWindow that this is indeed the main widget
    setCentralWidget(m_view);

    // then, setup our actions
    setupActions();

    // add a status bar
    statusBar()->show();
    
    setupDockers();

    // a call to KXmlGuiWindow::setupGUI() populates the GUI
    // with actions, using KXMLGUI.
    // It also applies the saved mainwindow settings, if any, and ask the
    // mainwindow to automatically save settings if changed: window size,
    // toolbar position, icon size, etc.
    setupGUI();
    
}

GTLDesigner::~GTLDesigner()
{
}

void GTLDesigner::setupDockers()
{
  // Error message
  QDockWidget* errorMessagesDockWidget = new QDockWidget(i18n("Error messages"), this);
  errorMessagesDockWidget->setObjectName( "ErrorMessagesDocker" );
  m_listView = new QListView( errorMessagesDockWidget );
  errorMessagesDockWidget->setWidget( m_listView );
  connect( m_listView, SIGNAL(activated(const QModelIndex &)),
            SLOT( errorLineActivated(const QModelIndex&)) );
  connect( m_listView, SIGNAL(clicked(const QModelIndex &)),
            SLOT( errorLineActivated(const QModelIndex&)) );
  addDockWidget( Qt::BottomDockWidgetArea, errorMessagesDockWidget );
    
  // Images docker
  QDockWidget* imagesDockerDockWidget = new QDockWidget(i18n("Images"), this);
  imagesDockerDockWidget->setObjectName("ImagesDocker");
  QWidget* imagesDockerWidget = new QWidget( imagesDockerDockWidget );
  imagesDockerDockWidget->setWidget( imagesDockerWidget );
  ui_imagesDocker = new Ui::ImagesDocker;
  ui_imagesDocker->setupUi( imagesDockerWidget );
  
  connect(ui_imagesDocker->buttonAdd, SIGNAL(released()), SLOT(addImage()));
  connect(ui_imagesDocker->buttonRemove, SIGNAL(released()), SLOT(removeImage()));
  
  addDockWidget( Qt::RightDockWidgetArea, imagesDockerDockWidget );
  
}

void GTLDesigner::setupActions()
{
    KStandardAction::openNew(this, SLOT(fileNew()), actionCollection());
    KStandardAction::open(this, SLOT(fileOpen()), actionCollection());
    KStandardAction::save(this, SLOT(fileSave()), actionCollection());
    KStandardAction::saveAs(this, SLOT(fileSaveAs()), actionCollection());
    KStandardAction::quit(qApp, SLOT(closeAllWindows()), actionCollection());

//     KStandardAction::preferences(this, SLOT(optionsPreferences()), actionCollection());

    // custom menu and menu item - the slot is in the class GTLDesignerView
    KAction *custom = new KAction( i18n("Configure editor"), this);
    connect(custom, SIGNAL(triggered()), SLOT(optionsEditorPreferences()));
    actionCollection()->addAction( QLatin1String("configure_editor"), custom );
    
    // Execute actions
    KAction* actionExecuteRun = new KAction(i18n("Run"), this );
    connect( actionExecuteRun, SIGNAL(triggered()), SLOT(executeRun()));
    actionCollection()->addAction( QLatin1String("execute_run"), actionExecuteRun );
}

void GTLDesigner::fileNew()
{
    // this slot is called whenever the File->New menu is selected,
    // the New shortcut is pressed (usually CTRL+N) or the New toolbar
    // button is clicked

    // create a new window
    (new GTLDesigner)->show();
}

void GTLDesigner::fileOpen()
{
  KUrl file = KFileDialog::getOpenFileName( KUrl(), "*.ctl *.shiva", this);
  openFile( file );
}

void GTLDesigner::openFile( const KUrl& file )
{
  m_document->openUrl(file);
  updateCaption();
  updateType();
}

void GTLDesigner::fileSave()
{
  m_document->documentSave();
  updateCaption();
  updateType();
}

void GTLDesigner::fileSaveAs()
{
  m_document->documentSaveAs();
  updateCaption();
  updateType();
}

void GTLDesigner::updateType()
{
  QString suffix = QFileInfo( m_document->url().fileName() ).suffix();
  if( suffix == "shiva" )
  {
    m_type = SHIVA_KERNEL;
  } else if( suffix == "ctl" )
  {
    m_type = CTL;
  } else {
    m_type = UNKNOWN;
  }
}

void GTLDesigner::executeRun()
{
  switch( m_type )
  {
    case SHIVA_KERNEL:
      executeShiva();
      break;
    case CTL:
      executeCTL();
      break;
    default:
      KMessageBox::error( this, "Unknow interpreter");
  }
}

void GTLDesigner::executeCTL()
{
  if( ui_imagesDocker->listImages->count() < 1 )
  {
    KMessageBox::error( this, "You need at least one image");
    return;
  }
  OpenCTL::Module m;
  m.setSource( "unamed", m_document->text().toAscii ().data() );
  m.compile();
  if(m.isCompiled())
  {
    GTLCore::String fileName =  ui_imagesDocker->listImages->item( 0 )->data( Qt::UserRole ).toString().toLatin1().data();
    const GTLImageIO::ImageDC* decoder = GTLImageIO::ImageDCRegistry::instance()->decoder( fileName );
    if( not decoder )
    {
      QMessageBox::warning(this, tr("GTLDesigner"),
                            tr("Cannot read image %1..")
                            .arg(fileName.c_str()));
      return;
    }
    GTLCore::RegionI region;
    GTLCore::AbstractImage* image = decoder->decode( fileName, &region, 0 );
    if( not image )
    {
      QMessageBox::warning(this, tr("GTLDesigner"),
                            tr("Cannot read image %1..")
                            .arg(fileName.c_str()));
      return;
    }
    OpenCTL::Program program( "apply", &m, image->pixelDescription() );
    if( program.isInitialised() )
    {
      program.apply( *image, *image );
      QImage res( region.columns(), region.rows(), QImage::Format_RGB32);
      for( int i = 0; i < region.columns(); ++i)
      {
        for( int j = 0; j < region.rows(); ++j)
        {
          quint8* px = image->data<quint8>(i, j );
          QRgb rgb = 0;
          switch( image->pixelDescription().channels() )
          {
            case 4:
              rgb = qRgba( px[0], px[1], px[2], px[3]);
              break;
            case 3:
              rgb = qRgb( px[0], px[1], px[2]);
              break;
            case 2:
              rgb = qRgba( px[0], px[0], px[0], px[1]);
              break;
            case 1:
              rgb = qRgb( px[0], px[0], px[0]);
              break;
          }
          res.setPixel( i, j, rgb );
        }
      }
      showImage( res);
    } else {
      KMessageBox::error( this, "Can't initialised the program, check that you have an 'apply' function");
    }
    delete image;
  } else {
    m_listView->setModel( new ErrorMessagesModel( m.compilationMessages().errors() ) );
  }
}

void GTLDesigner::executeShiva()
{
  int maxWidth = -1;
  int maxHeight = -1;
  std::list<const GTLCore::AbstractImage*> inputImages;
  std::list<GTLCore::RegionI> inputDOD;
  for( int k = 0; k < ui_imagesDocker->listImages->count(); ++k)
  {
    GTLCore::String fileName =  ui_imagesDocker->listImages->item( k )->data( Qt::UserRole ).toString().toLatin1().data();
    const GTLImageIO::ImageDC* decoder = GTLImageIO::ImageDCRegistry::instance()->decoder( fileName );
    if( not decoder )
    {
      QMessageBox::warning(this, tr("GTLDesigner"),
                            tr("Cannot read image %1..")
                            .arg(fileName.c_str()));
    }
    GTLCore::RegionI region;
    GTLCore::AbstractImage* image = decoder->decode( fileName, &region, 0 );
    if( not image )
    {
      QMessageBox::warning(this, tr("GTLDesigner"),
                            tr("Cannot read image %1..")
                            .arg(fileName.c_str()));
    }
    inputImages.push_back( image );
    maxWidth = qMax(region.columns(), maxWidth);
    maxHeight = qMax(region.rows(), maxHeight);
    inputDOD.push_back( region );
  }
  OpenShiva::Kernel* kernel = new OpenShiva::Kernel(4);
  if( maxHeight != -1 ) kernel->setParameter(OpenShiva::Kernel::IMAGE_HEIGHT, GTLCore::Value((float)maxHeight));
  if( maxWidth != -1 ) kernel->setParameter(OpenShiva::Kernel::IMAGE_WIDTH, GTLCore::Value((float)maxWidth));
  kernel->setSource(m_document->text().toAscii().data());
  kernel->compile();
  if(kernel->isCompiled())
  {
    m_listView->setModel( 0 );
    GTLCore::RegionF region(0,0,1,1);
    if( kernel->hasGeneratedFunction() )
    {
      region = kernel->generated();
    }
    if( kernel->hasChangedFunction() )
    {
      int i = 0;
      for( std::list<GTLCore::RegionI>::iterator it = inputDOD.begin(); it != inputDOD.end(); ++it, ++i)
      {
        region += kernel->changed( *it, i, inputDOD );
      }
    }
    if( region.columns() == 1 or region.rows() == 1 )
    {
      region = GTLCore::RegionF(0,0,800,600); // Use sane size
    }
    kDebug() << region.columns() << " " << region.rows();
    
    if(not m_shivaRenderWindow)
    {
      m_shivaRenderWindow = new ShivaRenderWindow;
    }
    m_shivaRenderWindow->show();
    m_shivaRenderWindow->setRegion(region.toRegionI());
    m_shivaRenderWindow->setImages(inputImages);
    m_shivaRenderWindow->setKernel(kernel);
  } else {
    m_listView->setModel( new ErrorMessagesModel( kernel->compilationMessages().errors() ) );
  }
}

void GTLDesigner::showImage(const QImage& _image)
{
    ImageViewer* viewer = new ImageViewer( );
    viewer->setImage(_image);
    viewer->show();
}

void GTLDesigner::optionsEditorPreferences()
{
  KTextEditor::EditorChooser::editor()->configDialog( this );
  KTextEditor::EditorChooser::editor()->writeConfig();
}

void GTLDesigner::optionsPreferences()
{
}

void GTLDesigner::updateCaption()
{
  QString caption = m_document->documentName();
  if( m_document->isModified())
  {
    caption += "*";
  }
  setCaption( caption );
}

void GTLDesigner::errorLineActivated( const QModelIndex & index )
{
  int line = m_listView->model()->data( index, ErrorMessagesModel::LineRole ).toInt();
  m_view->setCursorPosition( KTextEditor::Cursor( line - 1, 0 ) );
  m_view->setFocus( Qt::ActiveWindowFocusReason );
}

void GTLDesigner::addImage()
{
  KUrl file = KFileDialog::getOpenFileName( KUrl(), GTLImageIO::ImageDCRegistry::instance()->decodableFilter().c_str(), this);
  QString fileName = file.toLocalFile();
  if (fileName.isEmpty()) return;
  QListWidgetItem* lwi = new QListWidgetItem( QFileInfo(fileName).fileName() );
  lwi->setData(Qt::UserRole, fileName );
  ui_imagesDocker->listImages->addItem( lwi );
}

void GTLDesigner::removeImage()
{
  delete ui_imagesDocker->listImages->currentItem();
}

#include "gtldesigner.moc"
